#include <stdio.h>

/* YOUR COMPILER MUST SUPPORT UNSIGNED LONG LONG INT! */

#define SHA1_VER "v0.32b"

/***********************************************************************
* Crufty SHA-1 hash program by S.T.L.                                  *
*       As of v0.28b+, should be *completely* compliant with the SHA-1 *
*       standard as in FIPS PUB 180-1 (for byte-oriented data).        *
* DISCLAIMER:                                                          *
*       This program and its source code are provided "as is" and are  *
*       guaranteed only to take up space on your hard drive. Any other *
*       possible uses of this program and its source code are done     *
*       completely AT YOUR RISK.                                       *
* Copyright (C) 2000 S.T.L.                                            *
*       This program is free software; you can redistribute it and/or  *
*       modify it under the terms of the GNU General Public License as *
*       published by the Free Software Foundation; either version 2    *
*       of the License, or (at your option) any later version.         * 
*       This program is distributed in the hope that it will be        *
*       useful, but WITHOUT ANY WARRANTY; without even the implied     *
*       warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR        *
*       PURPOSE.  See the GNU General Public License for more details. *
*       You should have received a copy of the GNU General Public      *
*       License along with this program; if not, write to the Free     *
*       Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, *
*       MA  02111-1307  USA                                            *
* Also, when I say "word", I mean it as in the SHA-1 spec: 32 bits.    *
*                                                                      * 
* - S.T.L.                                                             *
***********************************************************************/

#define MASK32 0xFFFFFFFFUL
/*********************************************************************
* This mask nonsense is retained because while without it, this code *
* would still work in DJGPP, it wouldn't be portable to systems with *
* 64-bit unsigned long ints. Circular shifts would cause seemingly   *
* discarded bits to reappear. The C standard guarantees only that    *
* unsigned long ints are AT LEAST 32-bit; not that they ARE 32-bit.  *
* A BUG was fixed here. Addition modulo 2^32 is equivalent to        *
* addition followed by bitwise-and with 2^32 - 1.                    *
*********************************************************************/

#define HEX4(X) (unsigned int) ((X) >> 16) , (unsigned int) ((X) & 0xFFFF)
/* Used for 4-hexit output */

#define CS(X, N) ( ( ((X) << N) | ((X) >> (32 - N)) ) & MASK32)
/* CS means CircularShift */

unsigned long int f(int t, unsigned long int b, unsigned long int c, unsigned long int d);
unsigned long int getpad(FILE *fp, int x, int y);
void shamessage(int option);
void readhash(unsigned long int flags, unsigned long int a, unsigned long int b, \
        unsigned long int c, unsigned long int d, unsigned long int e);

/* Must be powers of 2 */
#define SHAFLAG_QUIET 1
#define SHAFLAG_EXQUIET 2
#define SHAFLAG_SILENT 4
#define SHAFLAG_HALF 8
#define SHAFLAG_HUGE 16
#define SHAFLAG_BIN 32
#define SHAFLAG_READ 64
#define SHAFLAG_VERIFY 128
#define SHAFLAG_DISAGREE 256

#define SHA_MESSAGE 1
#define SHA_BOTH 2
#define SHA_BOTH_END 3
#define SHA_COMPILE 4
#define SHA_VERSHIST 5
#define SHA_SWITCHES 6
#define SHA_USAGE 7

/* The possibilities for arguments is rather large. I believe */
/* that I can detect most illegal combinations, but some rather */
/* cruddy things can probably get by. Oh well. */

#define SHA_MAXARGS 9

int stopflag = 0; /* Flag that stops hashing */

int main(int argc, char *argv[]) {
        /* Various things I use: */

        FILE *sfp = NULL; /* Subject file pointer */
        FILE *bfp = NULL; /* Binary file pointer [binfile] */
        int i, j; /* Counter variables */
        int carg; /* Used for stepping through switches */
        unsigned long int shaflag = 0; /* Flags for switches */
        int readbyte; /* Used in reading of binary files */

        /* Various things SHA-1 uses: */

        unsigned long int k[80], w[80]; /* 80-word arrays, thanks AYL */
        unsigned long int a, b, c, d, e, h[5]; /* Two 5-word buffers */
        unsigned long int temp; /* Word buffer */
        int t = 0; /* t as defined in Section 7 of the SHA-1 spec */
        
        if ((argc == 1) || (argc > SHA_MAXARGS)) {
                shamessage(SHA_BOTH_END);
                exit(0);
        }

        /* Process all switches, they begin with '-' */

        /* --argc > 0 catches when there are no more args to process */
        /* (*++argv)[0] == '-' catches when the arg is not a switch */

        while( (--argc > 0) && ( (*++argv)[0] == '-') ) {
                while ((carg = *++argv[0])) {
                        switch (carg) {
                        /* Lone switches */
                        case 'C':
                                shamessage(SHA_COMPILE);
                                exit(0);
                        case 'S':
                                shamessage(SHA_SWITCHES);
                                exit(0);
                        case 'V':
                                shamessage(SHA_VERSHIST);
                                exit (0);
                        /* Normal switches */
                        case 'b':
                                if (shaflag & SHAFLAG_BIN) {
                                        shamessage(SHA_BOTH);
                                        printf("ERROR: Repeated -b switch.\n");
                                        exit(1);
                                }
                                if (shaflag & SHAFLAG_READ) {
                                        shamessage(SHA_BOTH);
                                        printf("ERROR: -b and -r switches used.\n");
                                        exit(1);
                                }
                                if (shaflag & SHAFLAG_VERIFY) {
                                        shamessage(SHA_BOTH);
                                        printf("ERROR: -b and -v switches used.\n");
                                        exit(1);
                                }
                                shaflag |= SHAFLAG_BIN;
                                break;
                        case 'h':
                                if (shaflag & SHAFLAG_HUGE) {
                                        shamessage(SHA_BOTH);
                                        printf("ERROR: Only two -h switches possible.\n");
                                        exit(1);
                                }        
                                if (shaflag & SHAFLAG_HALF) {
                                        shaflag |= SHAFLAG_HUGE;
                                        break;
                                }
                                shaflag |= SHAFLAG_HALF;
                                break;
                        case 'q':
                                if (shaflag & SHAFLAG_SILENT) {
                                        shamessage(SHA_BOTH);
                                        printf("ERROR: Only three -q switches possible.\n");
                                        exit(1);
                                }        
                                if (shaflag & SHAFLAG_EXQUIET) {
                                        shaflag |= SHAFLAG_SILENT;
                                        break;
                                }
                                if (shaflag & SHAFLAG_QUIET) {
                                        shaflag |= SHAFLAG_EXQUIET;
                                        break;
                                }
                                shaflag |= SHAFLAG_QUIET;
                                break;
                        case 'r':
                                if (shaflag & SHAFLAG_BIN) {
                                        shamessage(SHA_BOTH);
                                        printf("ERROR: -b and -r switches used.\n");
                                        exit(1);
                                }
                                if (shaflag & SHAFLAG_READ) {
                                        shamessage(SHA_BOTH);
                                        printf("ERROR: Repeated -r switch.\n");
                                        exit(1);
                                }        
                                if (shaflag & SHAFLAG_VERIFY) {
                                        shamessage(SHA_BOTH);
                                        printf("ERROR: -r and -v switches used.\n");
                                        exit(1);
                                }
                                shaflag |= SHAFLAG_READ;
                                break;
                        case 'v':
                                if (shaflag & SHAFLAG_BIN) {
                                        shamessage(SHA_BOTH);
                                        printf("ERROR: -b and -v switches used.\n");
                                        exit(1);
                                }
                                if (shaflag & SHAFLAG_READ) {
                                        shamessage(SHA_BOTH);
                                        printf("ERROR: -r and -v switches used.\n");
                                        exit(1);
                                }
                                if (shaflag & SHAFLAG_VERIFY) {
                                        shamessage(SHA_BOTH);
                                        printf("ERROR: Repeated -v switch.\n");
                                        exit(1);
                                }        
                                shaflag |= SHAFLAG_VERIFY;
                                break;
                        default:
                                shamessage(SHA_BOTH);
                                printf("ERROR: Illegal option %c\n", carg);
                                exit(1);
                        }
                }
        }

        /* Detect flag silliness */

        if ((shaflag & SHAFLAG_SILENT) && ( (shaflag & SHAFLAG_BIN) == 0)) {
                shamessage(SHA_BOTH);
                printf("ERROR: -qqq switch detected, but no -b switch.\n");
                printf("USAGE: Extremely silly.\n");
                exit(0);
        }
                
        /* argc is number of args left */
        /* *argv[0] should point to subject filename */
        /* If bin switch, argv[1] should point to binary output filename */
        /* If verify switch, argv[1] should point to binary input filename */
        /* If read switch, argv[0] should point to subject filename */

        if (((shaflag & SHAFLAG_BIN) == 0) && ((shaflag & SHAFLAG_VERIFY) == 0) && (argc != 1)) {
                shamessage(SHA_BOTH);
                printf("ERROR: Filename not found!\n");
                exit(1);
        } else if ((shaflag & SHAFLAG_BIN) && (argc != 2)) {
                shamessage(SHA_BOTH);
                printf("ERROR: -b flag detected, two filenames not found!\n");
                exit(1);
        } else if ((shaflag & SHAFLAG_VERIFY) && (argc != 2)) {
                shamessage(SHA_BOTH);
                printf("ERROR: -v flag detected, two filenames not found!\n");
                exit(1);
        }
               
        if ((shaflag & SHAFLAG_QUIET) == 0)
                shamessage(SHA_MESSAGE);

         /* Input filename, for any read, is *argv[0] */

        if ((sfp = fopen(*argv, "rb")) == NULL) {
                printf("ERROR: Cannot open %s.\n", *argv);
                exit(1);
        }
        
        /* If read switch, blast out file as hash */

        if (shaflag & SHAFLAG_READ) {
                
                /* Fill up five words */

                for(i = 0; i < 5; i++) {
                        for(j = 0, h[i] = 0; j < 4; j++) {
                                if( (readbyte = getc(sfp)) == EOF) {
                                        printf("ERROR: Read switch detected. But file length less than 20 bytes!\n");
                                        exit(1);
                                }
                                h[i] |= ((unsigned long int) readbyte) << (8 * (3 - j));
                        }
                }

                if (getc(sfp) != EOF) {
                        printf("ERROR: Read switch detected. But file length longer than 20 bytes!\n");
                        exit(1);
                }

                if ((shaflag & SHAFLAG_EXQUIET) == 0)
                        printf("Reading from %s for 160-bit SHA-1 hash.\n", *argv);
                
                readhash(shaflag, h[0], h[1], h[2], h[3], h[4]);
                exit(0);
        } /* End read flag stuff */

        /* If bin flag on, output filename is argv[1] */

        if (shaflag & SHAFLAG_BIN) {
                if ((bfp = fopen(argv[1], "rb")) != NULL) {
                        printf("ERROR: Output file %s already exists!\n", argv[1]);
                        exit(1);
                }
                if ((bfp = fopen(argv[1], "wb")) == NULL) {
                        printf("ERROR: Cannot open file %s for output!\n", argv[1]);
                        exit(1);
                }
        }

        /* If verify flag on, binary input filename is argv[1] */

        if (shaflag & SHAFLAG_VERIFY) {
                if ((bfp = fopen(argv[1], "rb")) == NULL) {
                        printf("ERROR: Cannot open %s.\n", argv[1]);
                        exit(1);
                }
        }

        /* If bin flag on, bfp now binary file for writing */
        /* If verify flag on, bfp now binary file for input */

        /* Now to initialize the k array and h buffer. */

        for(i = 0; i < 20; i++)
                k[i] = 0x5A827999UL;
        for(i = 20; i < 40; i++)
                k[i] = 0x6ED9EBA1UL;
        for(i = 40; i < 60; i++)
                k[i] = 0x8F1BBCDCUL;
        for(i = 60; i < 80; i++)
                k[i] = 0xCA62C1D6UL;

        h[0] = 0x67452301UL;
        h[1] = 0xEFCDAB89UL;
        h[2] = 0x98BADCFEUL;
        h[3] = 0x10325476UL;
        h[4] = 0xC3D2E1F0UL;

        /* Begin block processing: Section 7 of SHA-1 spec */

        while(stopflag == 0) {
                /* Step (a) in the spec */
                /* Read in the 16 words of the block as W[0]... W[15] */
                
                for(i = 0; i < 16; i++) 
                        for(j = 0, w[i] = 0; j < 4; j++)
                                w[i] |= getpad(sfp, i, j); /* Oh, yeah */
                
                /* Time for (b) */

                for(t = 16; t < 80; t++)
                        w[t] = CS(w[t-3] ^ w[t-8] ^ w[t-14] ^ w[t-16], 1);

                /* Time for (c) */

                a = h[0], b = h[1], c = h[2], d = h[3], e = h[4];

                /* Time for (d) */

                for(t = 0; t < 80; t++) {
                        temp = (CS(a, 5) + f(t, b, c, d) + e + w[t] + k[t]) & MASK32;
                        e = d, d = c, c = CS(b, 30), b = a, a = temp;
                }

                /* Time for (e) */

                h[0] = (h[0] + a) & MASK32;
                h[1] = (h[1] + b) & MASK32;
                h[2] = (h[2] + c) & MASK32;
                h[3] = (h[3] + d) & MASK32;
                h[4] = (h[4] + e) & MASK32;

        } /* End block processing */

        /* Write binary file, if needed */
        if (shaflag & SHAFLAG_BIN) {
                for(i = 0; i <= 4; i++) {
                        for(j = 24; j >= 0; j -= 8) 
                                putc( (h[i] >> j) & 0xFF, bfp);
                }
        } /* Binary file blasted. */

        /* Verify against binary file, if needed */
        /* Trashes w[] for easy storage */

        if (shaflag & SHAFLAG_VERIFY) {
               
                /* Fill up five words */

                for(i = 0; i < 5; i++) {
                        for(j = 0, w[i] = 0; j < 4; j++) {
                                if( (readbyte = getc(bfp)) == EOF) {
                                        printf("ERROR: Verify switch detected. But file length less than 20 bytes!\n");
                                        exit(1);
                                }
                                w[i] |= ((unsigned long int) readbyte) << (8 * (3 - j));
                        }
                }

                if (getc(bfp) != EOF) {
                        printf("ERROR: Verify switch detected. But file length longer than 20 bytes!\n");
                        exit(1);
                }

                if ((shaflag & SHAFLAG_EXQUIET) == 0)
                        printf("Reading from %s for 160-bit SHA-1 hash.\n", argv[1]);

                for(i = 0; i < 5; i++) {
                        if (w[i] != h[i]) {
                                printf("Hashes do not agree!\n");
                                shaflag |= SHAFLAG_DISAGREE;
                                break;
                        }
                }

                if (!(shaflag & SHAFLAG_DISAGREE))
                        printf("Hash of %s and hash stored in %s agree.\n", argv[0], argv[1]);

                if ((shaflag & SHAFLAG_EXQUIET) == 0) {
                        printf("Hash stored in %s:\n", argv[1]);
                        readhash(shaflag, w[0], w[1], w[2], w[3], w[4]);
                        printf("Hash computed from %s:\n", argv[0]);
                }
        } /* End verify flag stuff */

        readhash(shaflag, h[0], h[1], h[2], h[3], h[4]);
        exit(0);

}  /* End main */


/* Ft (B, C, D) as defined in the spec */

unsigned long int f(int t, unsigned long int b, unsigned long int c, unsigned long int d) {
        /* Useful to know: 0 <= t <= 80 ALWAYS, otherwise the program */
        /* has already screwed up, hence the simple If tests here */
        if (t <= 19)
                return ((b & c) | ((~b) & d));
        if ((t >= 40) && (t <= 59))
                return ((b & c) | (b & d) | (c & d));
        return (b ^ c ^ d); /* t is either between 20 and 39, or 60 and 79 */
}


/* Messages that were bulking up main() */

void shamessage(int option) {
        switch (option) {
        case SHA_MESSAGE:
                printf("SHA1.EXE %s by S.T.L.", SHA1_VER); /* No newline */
                printf("  This program is distributed under the GNU GPL.\n");
                printf("Copyright (C) 2000 S.T.L.\n");
                printf("This comes with ABSOLUTELY NO WARRANTY; see the GNU GPL for details.\n");
                printf("This is freedom software, and you are welcome to redistribute it\n");
                printf("under certain conditions; see the GNU GPL for details.\n");
                printf("\n");
                break;
        case SHA_BOTH:
                shamessage(SHA_MESSAGE);
                shamessage(SHA_USAGE);
                printf("\n");
                break;
        case SHA_BOTH_END:
                shamessage(SHA_MESSAGE);
                shamessage(SHA_USAGE);
                break;
        case SHA_USAGE:
                printf("Usage: sha1 [-b | -r | -v] [-h] [-q] filename [binfile]\n");
                printf("OR:    sha1 [-C | -S | -V]   (Lone switch usage)\n");
                printf("\n");
                printf("To see descriptions of possible switches, type:\n");
                printf("sha1 -S\n");
                break;
        case SHA_COMPILE:
                printf("Compilation instructions using DJGPP and Win 9x:\n");
                printf("        Simple compilation:\n");
                printf("                gcc -Wall -o sha1.exe sha1.c\n");
                printf("        Advanced optimized compilation (switches.txt is already provided):\n");
                printf("                gcc @switches.txt\n");
                printf("Assuming you're using the same DJGPP version I'm using (most recent when I\n");
                printf("compiled this) and then you use the Ultimate Packer for eXecutables, UPX,\n");
                printf("with the --best switch, you should get an identical binary to my version.\n");
                printf("With a little intelligence, that switches.txt can be used on MANY other\n");
                printf("programs.\n");
                printf("DJGPP home page: www.delorie.com\n");
                printf("UPX home page: upx.tsx.org\n");
                break;
        case SHA_VERSHIST:
                printf("Version history:\n");
                printf("[Prehistory of nonfunctional alpha versions]\n");
                printf("v0.01b was the first program that output a (corrupted) hash.\n");
                printf("Each time v0.01b was revised, version += .01.\n");
                printf("[Interlude: many bugs squashed]\n");
                printf("v0.17b was the first to correctly hash Appendices A, B, and C.\n");
                printf("v0.18b included this message and compilation instructions.\n");
                printf("v0.19b was the first version to be publicly released.\n");
                printf("v0.20b massively cleaned up some code.\n");
                printf("v0.21b got rid of the tempfile! F3 works in DOS again!\n");
                printf("v0.22b added some features (-h switch) and cleaned up code.\n");
                printf("v0.23b added some features again (-qq) and cleaned up code.\n");
                printf("v0.24b added the -b switch.\n");
                printf("v0.25b added the -r switch.\n");
                printf("v0.26b added unsigned long long ints!\n");
                printf("v0.27b cleaned up some code, added -hh switch.\n");
                printf("v0.28b attempted to fix a bug with MASK32.\n");
                printf("v0.29b added some more features.\n");
                printf("v0.30b added on-the-fly file length detection!\n");
                printf("v0.31b made some minor changes.\n");
                printf("v0.32b added the -v switch.\n");
                break;
        case SHA_SWITCHES:
                shamessage(SHA_USAGE);
                printf("\n");
                printf("Some switches may be multiplied, e.g. -qq or -qqq\n");
                printf("Switches may be combined, e.g. -hq or -bqhq or -hq -q\n");
                printf("Lone switches display info, making SHA1.EXE ignore any other switches.\n");
                printf("Switches:\n");
                printf("-b     Binary, outputs hash as a 20-byte binary file binfile.\n");
                printf("-h     Half-word, outputs 4-hexit chunks, e.g. DEAD BEEF DEAD BEEF.\n");
                printf("-hh    Huge, outputs 40 hexits with no spaces, e.g. DEADBEEFDEADBEEF...\n");
                printf("-q     Quiet, outputs only the hash.\n");
                printf("-qq    Superquiet, prints only the hash digits.\n");
                printf("-qqq   Silent, prints nothing at all. Only useful with -b.\n");
                printf("-r     Read, takes filename and prints the hash inside.\n");
                printf("-v     Verify, hashes filename and compares to the hash stored in binfile.\n");
                printf("Lone switches:\n");
                printf("-C     Compiling instructions, how to make SHA1.EXE from sha1.c\n");
                printf("-S     This message.\n");
                printf("-V     Version history, from v0.01b to today.\n");
                break;
        default:
                printf("*SHA1.EXE screams and dies in shamessage*\n");
                break;
        }
}        


/* Reads a hash of five words */

void readhash(unsigned long int flags, unsigned long int a, unsigned long int b, \
unsigned long int c, unsigned long int d, unsigned long int e) {
        if (flags & SHAFLAG_SILENT)
                exit(0); /* Exit before printing anything */

        if ((flags & SHAFLAG_EXQUIET) == 0) 
                printf("SHA-1 hexadecimal hash:\n");

        if (flags & SHAFLAG_HUGE) {
                printf("%08lX%08lX%08lX%08lX%08lX\n", a, b, c, d, e);
        } else if (flags & SHAFLAG_HALF) {
                printf("%04X %04X %04X %04X %04X %04X %04X %04X %04X %04X\n", \
                HEX4(a), HEX4(b), HEX4(c), HEX4(d), HEX4(e) );
        } else {
                printf("%08lX %08lX %08lX %08lX %08lX\n", a, b, c, d, e);
        }
}


/************************************************************
* The COOLEST function in this program. Why, you ask?       *
* It pads on the fly AND determines file length on the fly! *
* Duffman says, "Oh, yeah!"                                 *
************************************************************/

unsigned long int getpad(FILE *fp, int x, int y) {

        /* x is the word in w[x], y is the byte (0, 1, 2, or 3 last) */
        int getbyte;

        /* Statics; guaranteed to be inited to 0 at program start */
        static unsigned long long int origbits;
        static int eofswitch;

        /* If EOF not occurred yet, test if EOF and if not return byte */

        if (eofswitch == 0) {
                getbyte = getc(fp);
                if (getbyte != EOF) {
                        origbits += 8;
                        return (((unsigned long int) getbyte) << (8 * (3 - y)));
                } else {
                        /* EOF has just occurred */
                        eofswitch = 1;
                        return (128UL << (8 * (3 - y)));
                }
        }

        /***********************************************************
        * EOF has occured, 0x80 already returned. Now to blast out *
        * 0s and at last the length. Since EOF occured and 0x80    *
        * returned, all words before w[14] are 0. (The only time   *
        * we'll not be returning 0s now is if we're returning the  *
        * file length, and that's in w[14], w[15].)                *
        ***********************************************************/

        if (x < 14)
                return 0;

        /* x is 14 or 15, must see if we're starting at (14, 0) */
        /* If so, start blasting out origbits. */
        /* If not, continue blasting out 0s. */

        if ((x == 14) && (y == 0))
                eofswitch = 2;

        if (eofswitch == 1)  /* Not time for file length yet */
                return 0;

        /* Eofswitch == 2 now */
        /* Blast out origbits */

        if (x == 14) {
                switch (y) {
                case 0:
                        return (unsigned long int) ((origbits & 0xFF00000000000000U) >> 32);
                case 1:
                        return (unsigned long int) ((origbits & 0x00FF000000000000U) >> 32);
                case 2:
                        return (unsigned long int) ((origbits & 0x0000FF0000000000U) >> 32);
                case 3:
                        return (unsigned long int) ((origbits & 0x000000FF00000000U) >> 32);  /* No other cases */
                }
        } else if (x == 15) {
                switch (y) {
                case 0:
                        return (unsigned long int) (origbits & 0xFF000000U);
                case 1:
                        return (unsigned long int) (origbits & 0x00FF0000U);
                case 2:
                        return (unsigned long int) (origbits & 0x0000FF00U);
                case 3:
                        stopflag = 1; /* End of padded message! */
                        return (unsigned long int) (origbits & 0x000000FFU);
                }
        }
        /* No other cases */
        printf("*SHA1.EXE screams and dies in getpad*\n");
        exit(9);
}
